{
    open Tokens
    open Util
}
let anything = ['a'-'z''A'-'Z''0'-'9''\128'-'\255']+ | ['\000'-'\255']

let anything_but_close_math =
 ( [^'<']
 | '<' [^'/']
 | "</" [^'m']
 | "</m" [^'a']
 | "</ma" [^'t']
 | "</mat" [^'h']
 | "</math" [^'>']
 ) +

let anything_but_close_pre =
 ( [^'<']
 | '<' [^'/']
 | "</" [^'p']
 | "</p" [^'r']
 | "</pr" [^'e']
 | "</pre" [^'>']
 ) +

let anything_but_close_nowiki =
 ( [^'<']
 | '<' [^'/']
 | "</" [^'n']
 | "</n" [^'o']
 | "</no" [^'w']
 | "</now" [^'i']
 | "</nowi" [^'k']
 | "</nowik" [^'i']
 | "</nowiki" [^'>']
 ) +

let anything_but_close_comment =
 ( [^'-']
 | '-' [^'-']
 | "--" [^'>']
 ) +

let whitespace = [' ''\t']
let digit = ['0'-'9']
let hexdigit = ['0'-'9''a'-'f''A'-'F']
let alphanum = ['a'-'z''A'-'Z''0'-'9']
let alpha = ['a'-'z''A'-'Z']

let entity_named = "&" alphanum + ";"
let entity_dec = "&#" digit + ";"
let entity_hex = "&#x" hexdigit + ";"

let html_space = [' ''\t''\r''\n']
let html_space_opt = html_space *

let html_attr_unquoted = ['a'-'z' 'A'-'Z' '0'-'9' '_' ',' ':' '-'] +
let html_attr_arg = ('\'' [^'\''] * '\'' | '"' [^'"'] * '"' | html_attr_unquoted)  
let html_attr_name = alpha +
let html_attr  = html_attr_name html_space * "=" html_space * html_attr_arg
let html_attrs = (html_space + html_attr) *

let html_tag_name = alpha alphanum *

let html_opening_tag = "<"  html_tag_name html_attrs html_space_opt ">"
let html_closing_tag = "</" html_tag_name html_space_opt ">"
let html_closed_tag  = "<"  html_tag_name html_space_opt "/>"

let articlename = [' ''-''a'-'z''A'-'Z''0'-'9'':''_''+'',''.'' ''{''}''\128'-'\255']+

rule token = parse
    '\n'
      { NL }
  | '\r'
      { token lexbuf }
  | whitespace +
      { SP (Lexing.lexeme lexbuf) }
  | ['#''*'] +
      { LIST (Lexing.lexeme lexbuf) }
  | "="
      { EQ1 }
  | "=="
      { EQ2 }
  | "==="
      { EQ3 }
  | "===="
      { EQ4 }
  | "====="
      { EQ5 }
  | "======"
      { EQ6 }
  | "''"
      { Q2 }
  | "'''"
      { Q3 }
  | "'''''"
      { Q5 }
  | "----" '-' *
      { HR }
  | "<pre>" anything_but_close_pre "</pre>"
      { PRE (string_brange (Lexing.lexeme lexbuf) 5 6) }
  | "<nowiki>" anything_but_close_nowiki "</nowiki>"
      { LIT (string_brange (Lexing.lexeme lexbuf) 8 9) }
  | "<math>" anything_but_close_math "</math>"
      { MATH (string_brange (Lexing.lexeme lexbuf) 6 7) }
  | "<!--" anything_but_close_comment "-->"
      { token lexbuf }
  | "[[" articlename "|"
      { LINK (title_link_form (string_brange (Lexing.lexeme lexbuf) 2 1)) }
  | "[[" ['a'-'z']['a'-'z']  ":" articlename "]]"
      { let s = Lexing.lexeme lexbuf in LINK_INTERWIKI (String.sub s 2 2, string_brange s 5 2) }
  | "[[" articlename "]]"
      { let title = string_brange (Lexing.lexeme lexbuf) 2 2 in LINKED (title_link_form title, html_escape title) }
  | "]]"
      { LINK_CLOSE }
  | "{{CURRENTMONTH}}"
      { VAR_CURRENTMONTH }
  | "{{CURRENTDAY}}"
      { VAR_CURRENTDAY }
  | "{{CURRENTYEAR}}"
      { VAR_CURRENTYEAR }
  | "{{CURRENTDAYNAME}}"
      { VAR_CURRENTDAYNAME }
  | "{{CURRENTTIME}}"
      { VAR_CURRENTTIME }
  | "{{NUMBEROFARTICLES}}"
      { VAR_NUMBEROFARTICLES }
  | ":" +
      { COLONS (Lexing.lexeme lexbuf) }
  | "~~~"
      { T3 }
  | "~~~~"
      { T4 }
  | "RFC" " " ? digit +
      { let s = Lexing.lexeme lexbuf in 
        let s = string_brange s (if s.[3] = ' ' then 4 else 3) 0 in
        rfc_parse s }
  | "ISBN" " " ? ['0'-'9''X''-'] +
      { let s = Lexing.lexeme lexbuf in 
        let s = string_brange s (if s.[4] = ' ' then 5 else 4) 0 in
        isbn_parse s }
  | entity_named
      { let code = string_brange (Lexing.lexeme lexbuf) 1 1 in parse_ent_named code }
  | entity_dec
      { let code = string_brange (Lexing.lexeme lexbuf) 2 1 in parse_ent_dec code }
  | entity_hex
      { let code = string_brange (Lexing.lexeme lexbuf) 3 1 in parse_ent_hex code }
  | html_opening_tag
      { parse_html_opening_tag (Lexing.lexeme lexbuf) }
  | html_closing_tag
      { parse_html_closing_tag (Lexing.lexeme lexbuf) }
  | html_closed_tag
      { parse_html_closed_tag (Lexing.lexeme lexbuf) }
  | anything
      { LIT (html_escape (Lexing.lexeme lexbuf)) }
  | eof
      { EOF }
